import bz2
import tarfile
import typing

import fsspec
import msgspec

from betfair_parser.spec.streaming import stream_decode


def iter_file(fsspec_url: str):
    with fsspec.open(fsspec_url, compression="infer") as f:
        yield from iter_stream(f)


def iter_tar_file(tar_file: str, file_path: str):
    tf = tarfile.open(tar_file)
    assert file_path in tf.getnames(), f"Can't find `{file_path}` in {tf.getnames()[:5]}... "
    f = tf.extractfile(file_path)
    assert f is not None
    if file_path.endswith("bz2"):
        f = bz2.open(f)
    yield from iter_stream(f)  # type: ignore


def iter_stream(file_like: typing.BinaryIO):
    for line in file_like:
        try:
            data = stream_decode(line)
        except (msgspec.DecodeError, msgspec.ValidationError) as e:
            print("ERR", e)
            print(msgspec.json.decode(line))
            raise e
        yield data



from typing import Union

from msgspec.json import Decoder

from betfair_parser.spec.streaming.mcm import MCM
from betfair_parser.spec.streaming.ocm import OCM
from betfair_parser.spec.streaming.status import Connection, Status


_STREAM_MESSAGE = Union[Connection, Status, MCM, OCM]
_STREAM_DECODER = Decoder(_STREAM_MESSAGE)


def stream_decode(raw: bytes):
    return _STREAM_DECODER.decode(raw)



from typing import Literal, Optional

from betfair_parser.spec.common import BaseMessage
from betfair_parser.strenums import DocumentedEnum, doc


class StatusErrorCode(DocumentedEnum):
    # General errors not sent with id linking to specific request (as no request context)
    INVALID_INPUT = doc(
        "Failure code returned when an invalid input is provided (could not " "deserialize the message)"
    )
    TIMEOUT = doc("Failure code when a client times out (i.e. too slow sending data)")

    # Specific to authentication
    NO_APP_KEY = doc("Failure code returned when an application key is not found in the message")
    INVALID_APP_KEY = doc("Failure code returned when an invalid application key is received")
    NO_SESSION = doc("Failure code returned when a session token is not found in the message")
    INVALID_SESSION_INFORMATION = doc("Failure code returned when an invalid session token is received")
    NOT_AUTHORIZED = doc("Failure code returned when client is not authorized to perform the operation")
    MAX_CONNECTION_LIMIT_EXCEEDED = doc(
        "Failure code returned when a client tries to create more connections than allowed to"
    )
    TOO_MANY_REQUESTS = doc("Failure code returned when a client makes too many requests within a short time period")

    # Specific to subscription requests
    SUBSCRIPTION_LIMIT_EXCEEDED = doc(
        "Customer tried to subscribe to more markets than allowed to - set to 200 markets by default"
    )
    INVALID_CLOCK = doc(
        "Failure code returned when an invalid clock is provided on re-subscription "
        "(check initialClk / clk supplied)"
    )

    # General errors which may or may not be linked to specific request id
    UNEXPECTED_ERROR = doc("Failure code returned when an internal error occurred on the server")
    CONNECTION_FAILED = doc("Failure code used when the client / server connection is terminated")


class Connection(BaseMessage, tag_field="op", tag=str.lower, frozen=True):
    """Connection Message"""

    connection_id: str


class Status(BaseMessage, tag_field="op", tag=str.lower, frozen=True):
    """Status Message"""

    status_code: Literal["SUCCESS", "FAILURE"]
    connection_closed: bool
    id: Optional[int] = None
    connections_available: Optional[int] = None
    connection_id: Optional[str] = None
    error_code: Optional[StatusErrorCode] = None
    error_message: Optional[str] = None

    @property
    def is_error(self):
        return self.status_code != "SUCCESS"



from typing import List, Literal, Optional

from betfair_parser.spec.common import BaseMessage


class MatchedOrder(BaseMessage, array_like=True, frozen=True):
    price: float
    size: float


class UnmatchedOrder(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: str
    p: float
    s: float
    side: Literal["B", "L"]
    status: Literal["E", "EC"]
    pt: str
    ot: str
    pd: int
    rfo: Optional[str] = None
    rfs: Optional[str] = None
    rc: Optional[str] = None
    rac: Optional[str] = None
    md: Optional[int] = None
    cd: Optional[int] = None
    ld: Optional[int] = None
    avp: Optional[float] = None
    sm: Optional[float] = None
    sr: Optional[float] = None
    sl: Optional[float] = None
    sc: Optional[float] = None
    sv: Optional[float] = None


class StrategyMatched(BaseMessage, frozen=True):
    mb: Optional[List[MatchedOrder]] = []
    ml: Optional[List[MatchedOrder]] = []


class OrderChanges(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: int
    full_image: Optional[bool] = False
    hc: Optional[float] = None
    uo: Optional[List[UnmatchedOrder]] = []
    mb: Optional[List[MatchedOrder]] = []
    ml: Optional[List[MatchedOrder]] = []
    smc: Optional[dict[str, StrategyMatched]] = None


class OrderAccountChange(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: str
    account_id: Optional[int] = None
    full_image: Optional[bool] = False
    orc: List[OrderChanges] = []
    closed: Optional[bool] = None


class OCM(BaseMessage, tag_field="op", tag=str.lower, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: int
    clk: str
    pt: int
    oc: List[OrderAccountChange] = []
    initial_clk: Optional[str] = None
    status: Optional[int] = None
    conflate_ms: Optional[int] = None
    heartbeat_ms: Optional[int] = None
    ct: Optional[Literal["HEARTBEAT", "SUB_IMAGE"]] = None
    con: Optional[bool] = None



"""Shared definitions between the different API parts."""

from betfair_parser.spec.common.enums import *  # noqa
from betfair_parser.spec.common.messages import *  # noqa
from betfair_parser.spec.common.type_definitions import *  # noqa



import datetime
from typing import Annotated, Optional, Union

import msgspec

from betfair_parser.spec.common.messages import BaseMessage


# Type aliases with minimalistic validation. More would be great.

Date = Annotated[datetime.datetime, msgspec.Meta(title="Date", tz=True)]
SelectionId = Annotated[int, msgspec.Meta(title="SelectionId")]
Venue = Annotated[str, msgspec.Meta(title="Venue")]
MarketId = Annotated[str, msgspec.Meta(title="MarketId")]
Handicap = Annotated[float, msgspec.Meta(title="Handicap")]
EventId = Annotated[str, msgspec.Meta(title="EventId")]
EventTypeId = Annotated[int, msgspec.Meta(title="EventTypeId")]
CountryCode = Annotated[str, msgspec.Meta(title="CountryCode", min_length=2, max_length=3)]
ExchangeId = Annotated[str, msgspec.Meta(title="ExchangeId")]
CompetitionId = Annotated[str, msgspec.Meta(title="CompetitionId")]
Price = Annotated[float, msgspec.Meta(title="Price")]
Size = Annotated[float, msgspec.Meta(title="Size")]
BetId = Annotated[Union[str, int], msgspec.Meta(title="BetId")]
MatchId = Annotated[Union[str, int], msgspec.Meta(title="MatchId")]
CustomerRef = Annotated[Union[str, int], msgspec.Meta(title="CustomerRef")]
CustomerOrderRef = Annotated[Union[str, int], msgspec.Meta(title="CustomerOrderRef")]
CustomerStrategyRef = Annotated[Union[str, int], msgspec.Meta(title="CustomerStrategyRef")]


class TimeRange(BaseMessage, frozen=True):
    from_: Optional[Date] = msgspec.field(name="from", default=None)
    to: Optional[Date] = None



from typing import Annotated, List, Literal, Optional, Union

import msgspec

from betfair_parser.spec.betting.enums import MarketStatus, RunnerStatus
from betfair_parser.spec.betting.type_definitions import PriceLadderDescription
from betfair_parser.spec.common import BaseMessage, Date, EventTypeIdCode, RegulatorCode


class RunnerValues(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    tv: float  # Traded Volume
    ltp: float  # Last Traded Price
    spn: float  # Starting Price Near
    spf: float  # Starting Price Far


class Runner(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    sort_priority: int
    id: Union[int, str]
    name: Optional[str] = None
    hc: Optional[Union[float, str]] = None
    status: Optional[RunnerStatus] = None
    adjustment_factor: Optional[float] = None
    selection_id: Optional[str] = None
    bsp: Optional[Union[str, float]] = None
    removal_date: Optional[str] = None

    @property
    def handicap(self) -> str:
        return str(self.hc or 0.0)

    @property
    def runner_id(self) -> int:
        return int(self.selection_id or self.id)


class RunnerKeyLine(BaseMessage, frozen=True):
    id: int
    hc: int


class KeyLineDefinition(BaseMessage, frozen=True):
    kl: list[RunnerKeyLine]


class MarketDefinition(BaseMessage, kw_only=True, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    bsp_market: bool
    turn_in_play_enabled: bool
    persistence_enabled: bool
    market_base_rate: Optional[float]
    market_id: Optional[str] = ""
    market_name: Optional[str] = ""
    event_id: str
    event_type_id: str
    number_of_winners: int
    betting_type: str
    market_type: str
    market_time: str
    competition_id: Optional[str] = ""
    competition_name: Optional[str] = ""
    event_name: Optional[str] = ""
    suspend_time: str
    bsp_reconciled: bool
    complete: bool
    in_play: bool
    cross_matching: bool
    runners_voidable: bool
    number_of_active_runners: int
    bet_delay: int
    status: MarketStatus
    runners: List[Runner]
    regulators: List[RegulatorCode]
    name: Optional[str] = None
    open_date: Optional[str] = None
    timezone: Optional[str] = None
    venue: Optional[str] = None
    version: Optional[int] = None
    country_code: Optional[str] = None
    discount_allowed: Optional[bool] = None
    race_type: Optional[str] = None
    price_ladder_definition: Optional[Union[str, PriceLadderDescription]] = None
    settled_time: Optional[Date] = None
    key_line_definition: Optional[KeyLineDefinition] = None
    each_way_divisor: Optional[float] = None

    @property
    def event_type_name(self) -> str:
        return EventTypeIdCode(int(self.event_type_id)).name


class _PriceVolume(BaseMessage, array_like=True, frozen=True):
    price: float
    volume: float


class _LevelPriceVolume(BaseMessage, array_like=True, frozen=True):
    level: int
    price: float
    volume: float


AvailableToBack = Annotated[_PriceVolume, msgspec.Meta(title="AvailableToBack")]
AvailableToLay = Annotated[_PriceVolume, msgspec.Meta(title="AvailableToLay")]
BestAvailableToBack = Annotated[_LevelPriceVolume, msgspec.Meta(title="BestAvailableToBack")]
BestAvailableToLay = Annotated[_LevelPriceVolume, msgspec.Meta(title="BestAvailableToLay")]
BestDisplayAvailableToBack = Annotated[_LevelPriceVolume, msgspec.Meta(title="BestDisplayAvailableToBack")]
BestDisplayAvailableToLay = Annotated[_LevelPriceVolume, msgspec.Meta(title="BestDisplayAvailableToLay")]
Trade = Annotated[_PriceVolume, msgspec.Meta(title="Trade")]
StartingPriceBack = Annotated[_PriceVolume, msgspec.Meta(title="StartingPriceBack")]
StartingPriceLay = Annotated[_PriceVolume, msgspec.Meta(title="StartingPriceLay")]


class RunnerChange(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: Union[int, str]
    atb: Optional[List[AvailableToBack]] = []
    atl: Optional[List[AvailableToLay]] = []
    batb: Optional[List[BestAvailableToBack]] = []
    batl: Optional[List[BestAvailableToLay]] = []
    bdatb: Optional[List[BestDisplayAvailableToBack]] = []
    bdatl: Optional[List[BestDisplayAvailableToLay]] = []
    spb: Optional[List[StartingPriceBack]] = []  # Starting Price (Available To) Back
    spl: Optional[List[StartingPriceLay]] = []  # Starting Price (Available To) Lay
    spn: Optional[Union[float, str]] = None  # Starting Price Near
    spf: Optional[Union[int, float, str]] = None  # Starting Price Far
    trd: Optional[List[Trade]] = []
    ltp: Optional[float] = None
    tv: Optional[float] = None
    hc: Optional[float] = None


class MarketChange(BaseMessage, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    id: str
    market_definition: Optional[MarketDefinition] = None
    rc: List[RunnerChange] = []
    img: bool = False
    tv: Optional[float] = None
    con: Optional[bool] = None


class MCM(BaseMessage, tag_field="op", tag=str.lower, frozen=True):
    """
    https://docs.developer.betfair.com/display/1smk3cen4v3lu3yomq5qye0ni/Exchange+Stream+API
    """

    pt: int
    clk: Optional[str] = None
    id: Optional[int] = None
    initial_clk: Optional[str] = None
    market_definition: Optional[MarketDefinition] = None
    status: Optional[int] = None
    conflate_ms: Optional[int] = None
    heartbeat_ms: Optional[int] = None
    ct: Optional[Literal["HEARTBEAT", "SUB_IMAGE", "RESUB_DELTA"]] = None
    mc: List[MarketChange] = []
    con: Optional[bool] = None

    @property
    def is_heartbeat(self):
        return self.ct == "HEARTBEAT"

    @property
    def stream_unreliable(self):
        return self.status == 503
